/*
** UC: 21111 - Sistemas Operativos
** e-fólio B 2023-24 (mtxor.c)
**
** Aluno: 2300163 - Luís Pereira
*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sched.h>
#include <unistd.h>

// Definindo a macro para habilitar ou desabilitar os prints extras
#define print_test_ativo 0

// Estrutura do buffer
typedef struct {
    int *items;
    int count;
    int dimbuf;
    int total_items_produced;
    int total_items_consumed;
    int N;
} buffer;

buffer buf;
pthread_mutex_t mtx_buf = PTHREAD_MUTEX_INITIALIZER;

// Estrutura para passar dados para as tarefas consumidoras
typedef struct {
    int id;
    int block_size;
    int total_operations;
    int local_xor;
} consumer_args;

// Função do produtor (main)
void producer(int N) {
    srand(737);
    int items_produced = 0;

    while (items_produced < N) {
        pthread_mutex_lock(&mtx_buf);
        if (buf.count == buf.dimbuf) {
            pthread_mutex_unlock(&mtx_buf);
            sched_yield();
        } else {
            buf.items[buf.count++] = rand();
            items_produced++;
            buf.total_items_produced++;
            pthread_mutex_unlock(&mtx_buf);
        }
    }

    // imprimir fim produtor se print_test_ativo estiver ativado
    #if print_test_ativo
    printf("fim produtor: ncount=%d buf.count=%d\n", items_produced, buf.count);
    #endif

}

// Função das tarefas consumidoras
void *consumer(void *arg) {
    consumer_args *args = (consumer_args *)arg;
    args->local_xor = 0;
    args->total_operations = 0;

    while (1) {
        pthread_mutex_lock(&mtx_buf);
        if (buf.total_items_produced == buf.N && buf.total_items_consumed == buf.N) {
            pthread_mutex_unlock(&mtx_buf);
            break;
        }

        if (buf.count == 0) {
            pthread_mutex_unlock(&mtx_buf);
            sched_yield();
            continue;
        }

        int items_to_consume = (buf.count < args->block_size) ? buf.count : args->block_size;
        for (int i = 0; i < items_to_consume; i++) {
            args->local_xor ^= buf.items[--buf.count];
            args->total_operations++;
            buf.total_items_consumed++;
        }
        pthread_mutex_unlock(&mtx_buf);
    }

    printf("T%d: %d operações realizadas\n", args->id, args->total_operations);
    pthread_exit((void *)(intptr_t)args->local_xor);
}

int main(int argc, char *argv[]) {
    if (argc != 4) {
        fprintf(stderr, "Formato incorrecto. Input: %s dimbuf N nt\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    int dimbuf = atoi(argv[1]);
    int N = atoi(argv[2]);
    int nt = atoi(argv[3]);

    if (dimbuf <= 0 || N <= 0 || nt < 1) {
        fprintf(stderr, "Valores inválidos para os argumentos.\n");
        exit(EXIT_FAILURE);
    }

    printf("Cálculo XOR de sequência com %d itens, %d tarefas e buffer %d itens\n", N, nt, dimbuf);

    buf.items = (int *)malloc(dimbuf * sizeof(int));
    buf.count = 0;
    buf.dimbuf = dimbuf;
    buf.total_items_produced = 0;
    buf.total_items_consumed = 0;
    buf.N = N;

    pthread_t consumers[nt];
    consumer_args args[nt];

    for (int i = 0; i < nt; i++) {
        args[i].id = i;
        args[i].block_size = (dimbuf / nt == 0) ? 1 : dimbuf / nt;
        pthread_create(&consumers[i], NULL, consumer, &args[i]);
    }

    producer(N);

    int total_operations = 0;
    int final_xor = 0;
    for (int i = 0; i < nt; i++) {
        void *ret_val;
        pthread_join(consumers[i], &ret_val);
        final_xor ^= (int)(intptr_t)ret_val;
        total_operations += args[i].total_operations;
    }

    printf("Número total de operações efetuadas: %d\n", total_operations);
    printf("Resultado final do XOR: %d\n", final_xor);

    // Imprimir Resultado esperado se print_test_ativo estiver ativado
    #if print_test_ativo
    srand(737);
    int expected_result = 0;
    for (int i = 0; i < N; i++) {
        expected_result ^= rand();
    }
    printf("Resultado esperado: %d\n", expected_result);
    #endif

    free(buf.items);
    pthread_mutex_destroy(&mtx_buf);

    return 0;
}